<?php

namespace Drupal\config_sync;

use Drupal\config_provider\InMemoryStorage;
use Drupal\config_sync\Plugin\SyncConfigCollectorInterface;
use Drupal\Core\Config\ConfigManagerInterface;
use Drupal\Core\Config\StorageComparer;
use Drupal\Core\Config\StorageInterface;
use Drupal\Core\Extension\Extension;

/**
 * The ConfigSyncSnapshotter provides helper functions for taking snapshots of
 * extension-provided configuration.
 */
class ConfigSyncSnapshotter implements ConfigSyncSnapshotterInterface {

  /**
   * The app root for the current operation.
   *
   * @var string
   */
  protected $root;

  /**
   * The configuration collector.
   *
   * @var \Drupal\config_sync\Plugin\SyncConfigCollectorInterface
   */
  protected $configCollector;

  /**
   * The provider configuration storage.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $providerStorage;

  /**
   * The snapshot config storage for values from the extension storage.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $snapshotStorage;

  /**
   * The configuration manager.
   *
   * @var \Drupal\Core\Config\ConfigManagerInterface
   */
  protected $configManager;

  /**
   * Constructs a ConfigSyncSnapshotter object.
   *
   * @param string $root
   *   The app root.
   * @param \Drupal\config_provider\Plugin\SyncConfigCollectorInterface $config_collector
   *   The config collector.
   * @param \Drupal\Core\Config\StorageInterface $provider_storage
   *   The provider configuration storage.
   * @param \Drupal\Core\Config\StorageInterface $snapshot_storage
   *   The snapshot storage for the items from the extension storage.
   * @param \Drupal\Core\Config\ConfigManagerInterface $config_manager
   *   The configuration manager.
   */
  public function __construct($root, SyncConfigCollectorInterface $config_collector, StorageInterface $provider_storage, StorageInterface $snapshot_storage, ConfigManagerInterface $config_manager) {
    $this->root = $root;
    $this->configCollector = $config_collector;
    $this->providerStorage = $provider_storage;
    $this->snapshotStorage = $snapshot_storage;
    $this->configManager = $config_manager;
  }

  /**
   * {@inheritdoc}
   */
  public function refreshExtensionSnapshot($type, array $names, $mode) {
    $extensions = [];
    foreach ($names as $name) {
      $pathname = $this->drupalGetFilename($type, $name);
      $extensions[$name] = new Extension($this->root, $type, $pathname);
    }

    switch ($mode) {
      case ConfigSyncSnapshotterInterface::SNAPSHOT_MODE_INSTALL:
        // Take the snapshot in two stages.
        // First, add the unaltered configuration.
        $this->configCollector->addConfigForSnapshotting($extensions);
        $this->saveSnapshotChanges();
        // Now alter the previously added configuration.
        $this->configCollector->alterConfigForSnapshotting($extensions);
        $this->saveSnapshotChanges();
        return;
      case ConfigSyncSnapshotterInterface::SNAPSHOT_MODE_IMPORT:
        // No special handling for alter on update since the item will already have
        // been altered on install.
        $this->configCollector->addInstallableConfig($extensions);
        $this->saveSnapshotChanges();
        return;
    }
  }

  /**
   * Saves changes from the provider storage to the snapshot storage.
   */
  protected function saveSnapshotChanges() {
    // Set up a storage comparer.
    $storage_comparer = new StorageComparer(
      $this->providerStorage,
      $this->snapshotStorage,
      $this->configManager
    );
    $storage_comparer->createChangelist();

    $changelist = $storage_comparer->getChangelist();

    // Only add new or changed items.
    foreach (['create', 'update'] as $changelist_type) {
      if (isset($changelist[$changelist_type])) {
        foreach ($changelist[$changelist_type] as $item_name) {
          if ($provided_data = $this->providerStorage->read($item_name)) {
            // Snapshot the configuration item as provided by the extension.
            $this->snapshotStorage->write($item_name, $provided_data);
          }
        }
      }
    }

    // Snapshot collections as well.
    foreach ($this->providerStorage->getAllCollectionNames() as $collection) {
      $changelist = $storage_comparer->getChangelist(NULL, $collection);
      $this->providerStorage = $this->providerStorage->createCollection($collection);
      $snapshot_storage = $this->snapshotStorage->createCollection($collection);

      // Only add new or changed items.
      foreach (['create', 'update'] as $changelist_type) {
        if (isset($changelist[$changelist_type])) {
          foreach ($changelist[$changelist_type] as $item_name) {
            if ($provided_data = $this->providerStorage->read($item_name)) {
              // Snapshot the configuration item as provided by the extension.
              $snapshot_storage->write($item_name, $provided_data);
            }
          }
        }
      }

    }

  }

  /**
   * {@inheritdoc}
   */
  public function createSnapshot() {
    // Take the snapshot in two stages.
    // First, add the unaltered configuration.
    $this->configCollector->addConfigForSnapshotting();
    $this->configManager->createSnapshot($this->providerStorage, $this->snapshotStorage);
    // Now alter the previously added configuration.
    $this->configCollector->alterConfigForSnapshotting();
    // Don't call ::createSnapshot() because we don't want to remove what we
    // added in the first stage.
    $this->copyStorage($this->providerStorage, $this->snapshotStorage);
  }

  /**
   * Copies the content of one storage to another.
   *
   * @param \Drupal\Core\Config\StorageInterface $source_storage
   *   The storage to synchronize configuration from.
   * @param \Drupal\Core\Config\StorageInterface $snapshot_storage
   *   The storage to synchronize configuration to.
   */
  protected function copyStorage($source_storage, $snapshot_storage) {
    foreach ($source_storage->listAll() as $name) {
      $snapshot_storage->write($name, $source_storage->read($name));
    }
    // Copy collections as well.
    foreach ($source_storage->getAllCollectionNames() as $collection) {
      $source_collection = $source_storage->createCollection($collection);
      $snapshot_collection = $snapshot_storage->createCollection($collection);
      foreach ($source_collection->listAll() as $name) {
        $snapshot_collection->write($name, $source_collection->read($name));
      }
    }
  }

  /**
   * Wrapper for drupal_get_filename().
   *
   * @param $type
   *   The type of the item; one of 'core', 'profile', 'module', 'theme', or
   *   'theme_engine'.
   * @param $name
   *   The name of the item for which the path is requested. Ignored for
   *   $type 'core'.
   *
   * @return string
   *   The path to the requested item or an empty string if the item is not
   *   found.
   */
  protected function drupalGetFilename($type, $name) {
    return drupal_get_filename($type, $name);
  }

}
