<?php

namespace Drupal\config_sync\Plugin\ConfigFilter;

use Drupal\config_filter\Plugin\ConfigFilterBase;
use Drupal\config_merge\ConfigMerger;
use Drupal\config_provider\InMemoryStorage;
use Drupal\config_sync\ConfigSyncListerInterface;
use Drupal\Core\Config\ConfigManagerInterface;
use Drupal\Core\Config\StorageInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a sync filter that brings in updates from installed extensions.
 *
 * @ConfigFilter(
 *   id = "config_sync",
 *   label = "Config Sync",
 *   storages = {"config_distro.storage.distro"},
 *   weight = 10,
 *   deriver = "\Drupal\config_sync\Plugin\ConfigFilter\SyncFilterDeriver"
 * )
 */
class SyncFilter extends ConfigFilterBase implements ContainerFactoryPluginInterface {

  /**
   * The sync source configuration storage.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $syncSourceStorage;

  /**
   * Constructs a new SyncFilter.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param array $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Config\StorageInterface $snapshot_storage
   *   The snapshot extension storage.
   * @param \Drupal\Core\Config\StorageInterface $provider_storage
   *   The configuration provider storage.
   * @param \Drupal\Core\Config\StorageInterface $active_storage
   *   The active configuration storage.
   */
  public function __construct(array $configuration, $plugin_id, array $plugin_definition, ConfigSyncListerInterface $config_sync_lister, StorageInterface $snapshot_storage, StorageInterface $provider_storage, StorageInterface $active_storage) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->initSyncSourceStorage($config_sync_lister, $snapshot_storage, $provider_storage, $active_storage);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('config_sync.lister'),
      $container->get('config_sync.snapshot_storage'),
      $container->get('config_provider.storage'),
      $container->get('config.storage')
    );
  }

  /**
   * Initializes the sync source storage.
   *
   * @param \Drupal\Core\Config\StorageInterface $snapshot_storage
   *   The snapshot extension storage.
   * @param \Drupal\Core\Config\StorageInterface $provider_storage
   *   The configuration provider storage.
   * @param \Drupal\Core\Config\StorageInterface $active_storage
   *   The active configuration storage.
   */
  protected function initSyncSourceStorage(ConfigSyncListerInterface $config_sync_lister, StorageInterface $snapshot_storage, StorageInterface $provider_storage, StorageInterface $active_storage) {
    $changelists = $config_sync_lister->getExtensionChangelist($this->configuration['extension_type'], $this->configuration['extension_name']);
    $sync_source_storage = new InMemoryStorage();

    foreach ($changelists as $collection => $changelist) {
      // Ensure storages are using the specified collection.
      foreach (['snapshot', 'provider', 'active', 'sync_source'] as $storage_prefix) {
        if ($collection !== ${$storage_prefix . '_storage'}->getCollectionName()) {
          ${$storage_prefix . '_storage'} = ${$storage_prefix . '_storage'}->getCollection($collection);
        }
      }
      // Process changes.
      if (!empty($changelist['create'])) {
        // To create, we simply save the new item to the merge storage.
        foreach (array_keys($changelist['create']) as $item_name) {
          $sync_source_storage->write($item_name, $provider_storage->read($item_name));
        }
      }
      // Process update changes.
      if (!empty($changelist['update'])) {
        $config_merger = new ConfigMerger();
        // To update, we merge the value into that of the active storage.
        foreach (array_keys($changelist['update']) as $item_name) {
          $current = $provider_storage->read($item_name);
          $previous = $snapshot_storage->read($item_name);
          $active = $active_storage->read($item_name);
          $merged = $config_merger->mergeConfigItemStates($previous, $current, $active);

          $sync_source_storage->write($item_name, $merged);
        }
      }
    }

    $this->syncSourceStorage = $sync_source_storage;
  }

  /**
   * Reads from the sync source configuration.
   *
   * @param string $name
   *   The name of the configuration to read.
   * @param mixed $data
   *   The data to be filtered.
   *
   * @return mixed
   *   The data filtered or read from the sync source storage.
   */
  protected function syncSourceStorageRead($name, $data) {
    if ($sync = $this->syncSourceStorage->read($name)) {
      if (!isset($sync['uuid']) && !isset($sync['_core'])) {
        // Merge in uuid and _core while retaining the key order.
        $merged = array_replace($data, $sync);
        $sync = array_intersect_key($merged, array_flip(array_merge(array_keys($sync), ['uuid', '_core'])));
      }
      return $sync;
    }

    return $data;
  }

  /**
   * Reads multiple from the sync source storage.
   *
   * @param array $names
   *   The names of the configuration to read.
   * @param array $data
   *   The data to filter.
   *
   * @return array
   *   The new data.
   */
  protected function syncSourceStorageReadMultiple(array $names, array $data) {
    $filtered_data = [];
    foreach ($names as $name) {
      $filtered_data[$name] = $this->syncSourceStorageRead($name, $data[$name]);
    }

    return $filtered_data;
  }

  /**
   * {@inheritdoc}
   */
  public function filterRead($name, $data) {
    return $this->syncSourceStorageRead($name, $data);
  }

  /**
   * {@inheritdoc}
   */
  public function filterExists($name, $exists) {
    return $exists || $this->syncSourceStorage->exists($name);
  }

  /**
   * {@inheritdoc}
   */
  public function filterReadMultiple(array $names, array $data) {
    $sync_data = $this->syncSourceStorageReadMultiple($names, $data);

    // Return the data with merged in sync source data.
    return array_merge($data, $sync_data);
  }

  /**
   * {@inheritdoc}
   */
  public function filterListAll($prefix, array $data) {
    $sync_names = $this->syncSourceStorage->listAll($prefix);

    return array_unique(array_merge($data, $sync_names));
  }

  /**
   * {@inheritdoc}
   */
  public function filterCreateCollection($collection) {
    return new static($this->configuration, $this->pluginId, $this->pluginDefinition, $this->syncSourceStorage->createCollection($collection));
  }

  /**
   * {@inheritdoc}
   */
  public function filterGetAllCollectionNames(array $collections) {
    return array_merge($collections, $this->syncSourceStorage->getAllCollectionNames());
  }

}
